
#ifndef PAM_RADIUS_H
#define PAM_RADIUS_H

#define _GNU_SOURCE
#include <features.h>

#include <stdio.h>

#ifndef __USE_POSIX2
#define __USE_POSIX2
#endif /* __USE_POSIX2 */

#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <stdarg.h>
#include <utmp.h>
#include <time.h>
#include <netdb.h>

#include <netinet/in.h>
#include <rpcsvc/ypclnt.h>
#include <rpc/rpc.h>

#include <pwdb/radius.h>
#include <pwdb/pwdb_radius.h>

/******************************************************************/

#endif /* PAM_RADIUS_H */
